@@ -18,6 +18,11 @@
 #include "globus_xio_popen_driver.h"
 #include "version.h"
 #include <stdio.h>
+#include <sys/types.h>
+#include <sys/wait.h>
+
+#define USE_SOCKET_PAIR 1
+#define GLOBUS_L_XIO_POPEN_WAITPID_DELAY 500
 
 GlobusDebugDefine(GLOBUS_XIO_POPEN);
 
@@ -68,17 +73,21 @@ GlobusXIODefineModule(popen) =
  */
 typedef struct xio_l_popen_attr_s
 {
+    globus_bool_t                       ignore_program_errors;
     globus_bool_t                       use_blocking_io;
     globus_bool_t                       pass_env;
     char *                              program_name;
     char **                             argv;
     int                                 argc;
+    char **                             env;
+    int                                 env_count;
     globus_xio_popen_preexec_func_t     fork_cb;
 } xio_l_popen_attr_t;
 
 /* default attr */
 static const xio_l_popen_attr_t         xio_l_popen_attr_default =
 {
+    GLOBUS_FALSE,
     GLOBUS_FALSE,
     GLOBUS_FALSE,
     NULL,
@@ -94,18 +103,39 @@ typedef struct xio_l_popen_handle_s
 {
     globus_xio_system_file_handle_t     in_system;
     globus_xio_system_file_handle_t     out_system;
+    globus_xio_system_file_handle_t     err_system;
     globus_xio_system_file_t            infd;
     globus_xio_system_file_t            outfd;
+    globus_xio_system_file_t            errfd;    
     globus_bool_t                       use_blocking_io;
+    globus_bool_t                       ignore_program_errors;
     globus_mutex_t                      lock; /* only used to protect below */
     globus_off_t                        file_position;
     pid_t                               pid;
+    int                                 wait_count;
+    int                                 kill_state;
+    globus_bool_t                       canceled;
     globus_xio_popen_preexec_func_t     fork_cb;
+    globus_xio_operation_t              close_op;
 } xio_l_popen_handle_t;
 
+enum
+{
+    GLOBUS_L_XIO_POPEN_NONE,
+    GLOBUS_L_XIO_POPEN_TERM,
+    GLOBUS_L_XIO_POPEN_KILL
+};
+
+
 #define GlobusXIOPOpenPosition(handle)                                \
     globus_l_xio_popen_update_position(handle, 0, SEEK_CUR)
 
+static
+void
+globus_l_popen_waitpid(
+    xio_l_popen_handle_t *              handle,
+    int                                 opts);
+
 static
 globus_off_t
 globus_l_xio_popen_update_position(
@@ -193,6 +223,15 @@ globus_l_xio_popen_attr_copy(
         }
         attr->argv[i] = NULL;
     }
+    if(src_attr->env_count > 0)
+    {
+        attr->env = (char **)globus_calloc(attr->env_count+1, sizeof(char*));
+        for(i = 0; i < attr->env_count; i++)
+        {
+            attr->env[i] = strdup(src_attr->env[i]);
+        }
+        attr->env[i] = NULL;
+    }
     *dst = attr;
 
     GlobusXIOPOpenDebugExit();
@@ -212,6 +251,7 @@ globus_l_xio_popen_attr_cntl(
 {
     int                                 i;
     char **                             argv;
+    char **                             env;
     xio_l_popen_attr_t *                attr;
     globus_xio_popen_preexec_func_t     cb;
     GlobusXIOName(globus_l_xio_popen_attr_cntl);
@@ -236,6 +276,21 @@ globus_l_xio_popen_attr_cntl(
             attr->program_name = strdup(attr->argv[0]);
             break;
 
+        case GLOBUS_XIO_POPEN_SET_CHILD_ENV:
+            env = va_arg(ap, char **);
+            for(attr->env_count = 0; env[attr->env_count] != NULL;
+                attr->env_count++)
+            {
+            }
+            attr->env = calloc(attr->env_count + 1, sizeof(char *));
+            for(i = 0; i < attr->env_count; i++)
+            {
+                attr->env[i] = strdup(env[i]);
+            }
+            attr->env[i] = NULL;
+            break;
+
+
         case GLOBUS_XIO_POPEN_SET_PASS_ENV:
             attr->pass_env = va_arg(ap, globus_bool_t);
             break;
@@ -244,6 +299,10 @@ globus_l_xio_popen_attr_cntl(
             attr->use_blocking_io = va_arg(ap, globus_bool_t);
             break;
 
+        case GLOBUS_XIO_POPEN_SET_IGNORE_ERRORS:
+            attr->ignore_program_errors = va_arg(ap, globus_bool_t);
+            break;
+
         case GLOBUS_XIO_POPEN_SET_PREEXEC_FUNC:
             cb = va_arg(ap, globus_xio_popen_preexec_func_t);
             attr->fork_cb = (globus_xio_popen_preexec_func_t) cb;
@@ -278,6 +337,14 @@ globus_l_xio_popen_attr_destroy(
         }
         free(attr->argv);
     }
+    if(attr->env_count > 0)
+    {
+        for(i = 0; i < attr->env_count; i++)
+        {
+            free(attr->env[i]);
+        }
+        free(attr->env);
+    }
     if(attr->program_name != NULL)
     {
         free(attr->program_name);
@@ -340,10 +407,12 @@ globus_l_xio_popen_child(
     xio_l_popen_attr_t *                attr,
     const globus_xio_contact_t *        contact_info,
     int *                               infds,
-    int *                               outfds)
+    int *                               outfds,
+    int *                               errfds)
 {
     int                                 rc;
 
+#   if !defined(USE_SOCKET_PAIR)
     close(outfds[1]);
     close(infds[0]);
     rc = dup2(outfds[0], STDIN_FILENO);
@@ -351,18 +420,63 @@ globus_l_xio_popen_child(
     {
         goto error;
     }
+    close(outfds[0]);
     rc = dup2(infds[1], STDOUT_FILENO);
     if(rc < 0)
     {
         goto error;
     }
+    close(infds[1]);
+#   else
+    rc = dup2(outfds[1], STDIN_FILENO);
+    if(rc < 0)
+    {
+        close(outfds[0]);
+        close(outfds[1]);
+        goto error;
+    }
+    rc = dup2(infds[1], STDOUT_FILENO);
+    if(rc < 0)
+    {
+        close(infds[0]);
+        close(infds[1]);
+        goto error;
+    }
+    close(infds[0]);
+    close(infds[1]);
+#   endif
+    
+    if(!attr->ignore_program_errors)
+    {
+        rc = dup2(errfds[1], STDERR_FILENO);
+        if(rc < 0)
+        {
+            close(errfds[0]);
+            close(errfds[1]);
+            goto error;
+        }
+    }
+    close(errfds[0]);
+    close(errfds[1]);
+
+    /* if necessary, make optional 
+    setsid();
+    */
+
     if(attr->pass_env)
     {
         rc = execv(attr->program_name, attr->argv);
     }
     else
     {
-        char *                          env[] = {0};
+        char *                          l_env[] = {0};
+        char **                         env;
+
+        env = l_env;
+        if(attr->env != NULL)
+        {
+            env = attr->env;
+        }
         rc = execve(attr->program_name, attr->argv, env);
     }
 
@@ -370,6 +484,32 @@ error:
     exit(rc);
 }
 
+static
+globus_result_t
+globus_l_xio_popen_init_child_pipe(
+    int                                 fd,
+    globus_xio_system_file_handle_t *   out_system)
+{
+    globus_result_t                     result;
+    GlobusXIOName(globus_l_xio_popen_init_child_pipe);
+
+    fcntl(fd, F_SETFD, FD_CLOEXEC);
+
+    result = globus_xio_system_file_init(out_system, fd);
+    if(result != GLOBUS_SUCCESS)
+    {
+        result = GlobusXIOErrorWrapFailed(
+            "globus_xio_system_file_init", result);
+        goto error_init;
+    }
+
+    return GLOBUS_SUCCESS;
+
+error_init:
+
+    return result;
+}
+
 /*
  *  open a file
  */
@@ -382,8 +522,10 @@ globus_l_xio_popen_open(
     globus_xio_operation_t              op)
 {
     int                                 rc;
+    int                                 s_fds[2];
     int                                 infds[2];
     int                                 outfds[2];
+    int                                 errfds[2];
     xio_l_popen_handle_t *              handle;
     xio_l_popen_attr_t *                attr;
     globus_result_t                     result;
@@ -393,6 +535,16 @@ globus_l_xio_popen_open(
     
     attr = (xio_l_popen_attr_t *) 
         driver_attr ? driver_attr : &xio_l_popen_attr_default;
+
+    /* check that program exists and is exec=able first */
+
+    rc = access(attr->program_name, R_OK | X_OK);
+    if(rc != 0)
+    {
+        result = GlobusXIOErrorSystemError("access check", errno);
+        goto error_handle;
+    }
+
     result = globus_l_xio_popen_handle_init(&handle);
     if(result != GLOBUS_SUCCESS)
     {
@@ -401,58 +553,130 @@ globus_l_xio_popen_open(
         goto error_handle;
     }
 
-    rc = pipe(infds);
-    if(rc != 0)
-    {
-        result = GlobusXIOErrorSystemError("pipe", errno);
-        goto error_in_pipe;
-    }
-    rc = pipe(outfds);
-    if(rc != 0)
-    {
-        result = GlobusXIOErrorSystemError("pipe", errno);
-        goto error_out_pipe;
-    }
-
-    handle->pid = fork();
-    if(handle->pid == 0)
-    {
-        globus_l_xio_popen_child(attr, contact_info, infds, outfds);
-    }
-    fcntl(outfds[1], F_SETFD, FD_CLOEXEC);
-    fcntl(infds[0], F_SETFD, FD_CLOEXEC);
-
-    handle->infd = infds[0];
-    handle->outfd = outfds[1];
-
-    result = globus_xio_system_file_init(&handle->out_system, handle->outfd);
-    if(result != GLOBUS_SUCCESS)
-    {
-        result = GlobusXIOErrorWrapFailed(
-            "globus_xio_system_file_init", result);
-        goto error_init;
-    }
-    result = globus_xio_system_file_init(&handle->in_system, handle->infd);
-    if(result != GLOBUS_SUCCESS)
-    {
-        result = GlobusXIOErrorWrapFailed(
-            "globus_xio_system_file_init", result);
-        goto error_init;
-    }
+    handle->ignore_program_errors = attr->ignore_program_errors;
+    handle->use_blocking_io = attr->use_blocking_io;
     
-    close(outfds[0]);
-    close(infds[1]);
+#   if defined(USE_SOCKET_PAIR)
+    {
+        rc = socketpair(AF_UNIX, SOCK_STREAM, 0, s_fds);
+        if(rc != 0)
+        {
+            result = GlobusXIOErrorSystemError("socketpair", errno);
+            goto error_in_pipe;
+        }
+
+        /* trick the rest of the code into thinking it is using pipe */
+        outfds[0] = s_fds[0];
+        outfds[1] = s_fds[1];
+        infds[0] = s_fds[0];
+        infds[1] = s_fds[1];
+    }
+#   else
+    {
+        rc = pipe(infds);
+        if(rc != 0)
+        {
+            result = GlobusXIOErrorSystemError("pipe", errno);
+            goto error_in_pipe;
+        }
+        rc = pipe(outfds);
+        if(rc != 0)
+        {
+            result = GlobusXIOErrorSystemError("pipe", errno);
+            goto error_out_pipe;
+        }
+    }
+#   endif
+
+    rc = pipe(errfds);
+    if(rc != 0)
+    {
+        result = GlobusXIOErrorSystemError("pipe", errno);
+        goto error_err_pipe;
+    }
+    fcntl(errfds[0], F_SETFL, O_NONBLOCK);
+    fcntl(errfds[1], F_SETFL, O_NONBLOCK);
+    
+    handle->pid = fork();
+    if(handle->pid < 0)
+    {
+        result = GlobusXIOErrorSystemError("fork", errno);
+        goto error_fork;
+    }
+    else if(handle->pid == 0)
+    {
+        globus_l_xio_popen_child(attr, contact_info, infds, outfds, errfds);
+    }
+
+#   if defined(USE_SOCKET_PAIR)
+    {
+        handle->infd = s_fds[0];
+        handle->outfd = s_fds[0];
+        
+        result = globus_l_xio_popen_init_child_pipe(
+            handle->infd,
+            &handle->in_system);
+        if(result != GLOBUS_SUCCESS)
+        {
+            goto error_init;
+        }
+        handle->out_system = handle->in_system;
+        close(s_fds[1]);
+    }
+#   else
+    {
+        handle->infd = infds[0];
+        handle->outfd = outfds[1];
+
+        result = globus_l_xio_popen_init_child_pipe(
+            handle->outfd,
+            &handle->out_system);
+        if(result != GLOBUS_SUCCESS)
+        {
+            goto error_init;
+        }
+        result = globus_l_xio_popen_init_child_pipe(
+            handle->infd,
+            &handle->in_system);
+        if(result != GLOBUS_SUCCESS)
+        {
+            goto error_init;
+        }
+        close(outfds[0]);
+        close(infds[1]);
+    }
+#   endif
+
+    handle->errfd = errfds[0];
+    result = globus_l_xio_popen_init_child_pipe(
+        handle->errfd,
+        &handle->err_system);
+    if(result != GLOBUS_SUCCESS)
+    {
+        goto error_init;
+    }
+    close(errfds[1]);
     globus_xio_driver_finished_open(handle, op, GLOBUS_SUCCESS);
     
     GlobusXIOPOpenDebugExit();
     return GLOBUS_SUCCESS;
 
 error_init:
+error_fork:
+    close(errfds[0]);
+    close(errfds[1]);
+#   if defined(USE_SOCKET_PAIR)
+error_err_pipe:
+    close(s_fds[0]);
+    close(s_fds[1]);
+#   else
+error_err_pipe:
     close(outfds[0]);
     close(outfds[1]);
 error_out_pipe:
     close(infds[0]);
     close(infds[1]);
+#   endif
 error_in_pipe:
     globus_l_xio_popen_handle_destroy(handle);
 error_handle:
@@ -460,6 +684,142 @@ error_handle:
     return result;
 }
 
+static
+void
+globus_l_xio_popen_close_oneshot(
+    void *                              arg)
+{
+    xio_l_popen_handle_t *              handle;
+    GlobusXIOName(globus_l_xio_popen_close_oneshot);
+
+    GlobusXIOPOpenDebugEnter();
+    handle = (xio_l_popen_handle_t *) arg;
+            
+    globus_l_popen_waitpid(handle, WNOHANG);
+
+    GlobusXIOPOpenDebugExit();
+}
+
+static
+void
+globus_l_popen_waitpid(
+    xio_l_popen_handle_t *              handle,
+    int                                 opts)
+{
+    globus_result_t                     result = GLOBUS_SUCCESS;
+    int                                 status;
+    int                                 rc;
+    globus_reltime_t                    delay;
+    GlobusXIOName(globus_l_popen_waitpid);
+
+    rc = waitpid(handle->pid, &status, opts);
+    if(rc > 0)
+    {
+        /* if program exited normally, but with a nonzero code OR
+         * program exited by signal, and we didn't signal it */
+        if(((WIFEXITED(status) && WEXITSTATUS(status) != 0) || 
+            (WIFSIGNALED(status) && handle->kill_state != GLOBUS_L_XIO_POPEN_NONE))
+            && !handle->ignore_program_errors)
+        {
+            /* read programs stderr and dump it to an error result */
+            globus_size_t                   nbytes = 0;
+            globus_xio_iovec_t              iovec;
+            char                            buf[8192];
+
+            iovec.iov_base = buf;
+            iovec.iov_len = sizeof(buf) - 1;
+            
+            result = globus_xio_system_file_read(
+                handle->err_system, 0, &iovec, 1, 0, &nbytes);
+            
+            buf[nbytes] = 0;
+
+            if(WIFEXITED(status))
+            {
+                result = globus_error_put(
+                    globus_error_construct_error(
+                        GLOBUS_XIO_MODULE,
+                        GLOBUS_NULL,
+                        GLOBUS_XIO_ERROR_SYSTEM_ERROR,
+                        __FILE__,
+                        _xio_name,
+                        __LINE__,
+                        _XIOSL("popened program exited with an error "
+                               "(exit code: %d):\n%s"),
+                        WEXITSTATUS(status), 
+                        buf));
+            }
+            else
+            {
+                result = globus_error_put(
+                    globus_error_construct_error(
+                        GLOBUS_XIO_MODULE,
+                        GLOBUS_NULL,
+                        GLOBUS_XIO_ERROR_SYSTEM_ERROR,
+                        __FILE__,
+                        _xio_name,
+                        __LINE__,
+                        _XIOSL("popened program was terminated by a signal"
+                               "(sig: %d)"),
+                        WTERMSIG(status)));
+            }
+        }
+
+        globus_xio_system_file_close(handle->errfd);
+        globus_xio_system_file_destroy(handle->err_system);
+
+        globus_xio_driver_finished_close(handle->close_op, result);
+        globus_l_xio_popen_handle_destroy(handle);
+    }
+    else if(rc < 0 || opts == 0)
+    {
+        result = GlobusXIOErrorSystemError("waitpid", errno);
+
+        globus_xio_system_file_close(handle->errfd);
+        globus_xio_system_file_destroy(handle->err_system);
+
+        globus_xio_driver_finished_close(handle->close_op, result);
+        globus_l_xio_popen_handle_destroy(handle);
+    }
+    else
+    {
+        
+        handle->wait_count++;
+        
+        if(handle->canceled)
+        {
+            switch(handle->kill_state)
+            {
+                case GLOBUS_L_XIO_POPEN_NONE:
+                    if(handle->wait_count > 5000 / GLOBUS_L_XIO_POPEN_WAITPID_DELAY)
+                    {
+                        handle->kill_state = GLOBUS_L_XIO_POPEN_TERM;
+                        kill(handle->pid, SIGTERM);
+                    }
+                    break;
+                case GLOBUS_L_XIO_POPEN_TERM:
+                    if(handle->wait_count > 15000 / GLOBUS_L_XIO_POPEN_WAITPID_DELAY)
+                    {
+                        handle->kill_state = GLOBUS_L_XIO_POPEN_KILL;
+                        kill(handle->pid, SIGKILL);
+                    }
+                    break;
+                default:
+                    break;
+            } 
+        }
+        
+        GlobusTimeReltimeSet(delay, 0, GLOBUS_L_XIO_POPEN_WAITPID_DELAY);
+        globus_callback_register_oneshot(
+            NULL,
+            &delay,
+            globus_l_xio_popen_close_oneshot,
+            handle);         
+    }
+
+    GlobusXIOPOpenDebugExit();
+}
+
 /*
  *  close a file
  */
@@ -476,16 +836,26 @@ globus_l_xio_popen_close(
     GlobusXIOPOpenDebugEnter();
     
     handle = (xio_l_popen_handle_t *) driver_specific_handle;
-    
+
+    handle->close_op = op;
     globus_xio_system_file_destroy(handle->in_system);
-    globus_xio_system_file_destroy(handle->out_system);
-    
+   
     globus_xio_system_file_close(handle->infd);
+
+#if !defined(USE_SOCKET_PAIR)
     globus_xio_system_file_close(handle->outfd);
-    
-    globus_xio_driver_finished_close(op, GLOBUS_SUCCESS);
-    globus_l_xio_popen_handle_destroy(handle);
-    
+    globus_xio_system_file_destroy(handle->out_system);
+#endif
+
+    if(globus_xio_driver_operation_is_blocking(op))
+    {
+        globus_l_popen_waitpid(handle, 0);
+    }
+    else
+    {
+        globus_l_popen_waitpid(handle, WNOHANG);
+    }
+
     GlobusXIOPOpenDebugExit();
     return GLOBUS_SUCCESS;
 }
@@ -498,14 +868,20 @@ globus_l_xio_popen_system_read_cb(
     void *                              user_arg)
 {
     globus_xio_operation_t              op;
+    xio_l_popen_handle_t *              handle;
     GlobusXIOName(globus_l_xio_popen_system_read_cb);
     
     GlobusXIOPOpenDebugEnter();
     
     op = (globus_xio_operation_t) user_arg;
     
+    handle = (xio_l_popen_handle_t *) 
+        globus_xio_operation_get_driver_specific(op);
+        
+    handle->canceled = globus_xio_operation_is_canceled(op);
+    
     globus_l_xio_popen_update_position(
-        (xio_l_popen_handle_t *) globus_xio_operation_get_driver_specific(op),
+        handle,
         nbytes,
         SEEK_CUR);
         
@@ -580,14 +956,20 @@ globus_l_xio_popen_system_write_cb(
     void *                              user_arg)
 {
     globus_xio_operation_t              op;
+    xio_l_popen_handle_t *              handle;
     GlobusXIOName(globus_l_xio_popen_system_write_cb);
     
     GlobusXIOPOpenDebugEnter();
     
     op = (globus_xio_operation_t) user_arg;
     
+    handle = (xio_l_popen_handle_t *) 
+        globus_xio_operation_get_driver_specific(op);
+        
+    handle->canceled = globus_xio_operation_is_canceled(op);
+
     globus_l_xio_popen_update_position(
-        (xio_l_popen_handle_t *) globus_xio_operation_get_driver_specific(op),
+        handle,
         nbytes,
         SEEK_CUR);
         
@@ -663,6 +1045,10 @@ static globus_xio_string_cntl_table_t popen_l_string_opts_table[] =
         globus_xio_string_cntl_bool},
     {"argv", GLOBUS_XIO_POPEN_SET_PROGRAM,
         globus_xio_string_cntl_string_list},
+    {"env", GLOBUS_XIO_POPEN_SET_CHILD_ENV,
+        globus_xio_string_cntl_string_list},
+    {"ignore_errors", GLOBUS_XIO_POPEN_SET_IGNORE_ERRORS,
+        globus_xio_string_cntl_bool},
     {0}
 };
 
